@@ -11,6 +11,7 @@ import threading
 import time
 import traceback
 from getpass import getpass
+from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
 from io import BytesIO
 from multiprocessing.dummy import Pool
 from pathlib import Path
@@ -26,11 +27,11 @@ from pydicom.encaps import encapsulate, generate_pixel_data_frame
 from pydicom.errors import InvalidDicomError
 from pydicom.pixel_data_handlers import apply_color_lut
 from pynetdicom import AE, evt, ALL_TRANSFER_SYNTAXES
-from pynetdicom.sop_class import UltrasoundMultiframeImageStorage, UltrasoundImageStorage
+from pynetdicom import sop_class
 from tqdm import tqdm
 from watchdog.events import FileCreatedEvent, FileMovedEvent, FileModifiedEvent
 from watchdog.events import FileSystemEventHandler
-from watchdog.observers import Observer
+from watchdog.observers.polling import PollingObserver
 
 from echoloader.lib import unpack
 from echoloader.sync import PacsConnection, Sync, DEFAULT_AE_TITLE
@@ -193,7 +194,7 @@ def pad_to_multiple(arr, size, dims=(1, 2)):
     return np.pad(arr, pad_dims, 'constant')
 
 
-def compress_dicom(ds, anonymize):
+def package_dicom(ds, anonymize, compress):
     # Populate required values for file meta information
     ds.remove_private_tags()
     media = parse_dicom_pixel(ds)
@@ -211,7 +212,7 @@ def compress_dicom(ds, anonymize):
     if len(media.shape) < 3 + video:
         media = np.repeat(np.expand_dims(media, -1), 3, -1)
     ds.Rows, ds.Columns, ds.SamplesPerPixel = media.shape[video:]
-    ds.PhotometricInterpretation = "RGB"
+    ds.PhotometricInterpretation = "YBR_FULL_422"
     if video:
         ds.StartTrim = 1
         ds.StopTrim = ds.NumberOfFrames = media.shape[0] if video else 1
@@ -221,9 +222,13 @@ def compress_dicom(ds, anonymize):
         ds.ActualFrameDuration = math.ceil(1000 / ds.CineRate)
         ds.PreferredPlaybackSequencing = 0
         ds.FrameDelay = 0
-        ds.file_meta.TransferSyntaxUID = uid.MPEG4HP41
-        ds.PixelData = encapsulate([
-            imageio.mimwrite(imageio.RETURN_BYTES, pad_to_multiple(media, 16), fps=fps, format='mp4')])
+        if compress:
+            ds.file_meta.TransferSyntaxUID = uid.MPEG4HP41
+            ds.PixelData = encapsulate([
+                imageio.mimwrite(imageio.RETURN_BYTES, pad_to_multiple(media, 16), fps=fps, format='mp4')])
+        else:
+            ds.file_meta.TransferSyntaxUID = uid.JPEGBaseline8Bit
+            ds.PixelData = encapsulate([imageio.imwrite(imageio.RETURN_BYTES, img, format='jpg') for img in media])
     else:
         ds.file_meta.TransferSyntaxUID = uid.JPEGBaseline8Bit
         ds.PixelData = encapsulate([imageio.imwrite(imageio.RETURN_BYTES, media, format='jpg')])
@@ -239,6 +244,39 @@ def wait_file(path):
         old, cur = cur, os.path.getsize(path)
 
 
+def create_proxy(args):
+    class Proxy(SimpleHTTPRequestHandler):
+        def do_request(self, method):
+            url = f'{args.auth.api_url}{self.path}'
+            headers = dict(self.headers)
+            headers.update(args.auth.get_headers())
+            headers['Host'] = 'proxy.us2.ai'
+            data = self.rfile.read(int(self.headers['Content-Length'] or 0))
+            response = getattr(requests, method)(url, headers=headers, stream=True, data=data)
+            self.send_response(response.status_code)
+            for k, v in response.headers.items():
+                self.send_header(k, v)
+            self.end_headers()
+            self.copyfile(response.raw, self.wfile)
+
+        def do_GET(self):
+            self.do_request('get')
+
+        def do_PUT(self):
+            self.do_request('put')
+
+        def do_POST(self):
+            self.do_request('post')
+
+        def do_DELETE(self):
+            self.do_request('delete')
+
+        def do_OPTIONS(self):
+            self.do_request('options')
+
+    return Proxy
+
+
 class Handler(FileSystemEventHandler):
     def __init__(self, args):
         self.args = args
@@ -261,14 +299,15 @@ class Handler(FileSystemEventHandler):
     def join(self):
         self.pool.join()
 
-    def file_params(self, ds, ae_title):
+    def file_params(self, ds, ae_title, filename=None):
         customer = getattr(self.args, 'customer', '') or ae_title
+        stem = '_'.join(map(str, filter(bool, [ds.SOPInstanceUID, filename])))
         return {
-            'customer': customer,
-            'trial': customer,
+            'customer': customer or '',
+            'trial': getattr(self.args, 'trial', None) or customer or '',
             'patient_id': ds.PatientID,
             'visit_id': ds.StudyID or ds.StudyInstanceUID or ds.StudyDate or 'No Study ID',
-            'filename': f"{ds.SOPInstanceUID}.dcm",
+            'filename': f"{stem}.dcm",
         }
 
     def upload(self, ds, param):
@@ -296,12 +335,17 @@ class Handler(FileSystemEventHandler):
         if self.closed:
             return
         path = path and Path(path)
-        params = self.file_params(ds or pydicom.dcmread(path, stop_before_pixels=True), ae_title=ae_title)
+        params = self.file_params(ds or pydicom.dcmread(path, stop_before_pixels=True), ae_title=ae_title,
+                                  filename=getattr(path, 'stem', None))
         extracted = self.args.extracted
         k = tuple(params.values())
         if k not in extracted:
             ds = ds or pydicom.dcmread(path)
-            compress_dicom(ds, anonymize=self.args.anonymize)
+            if self.args.filter and not any(
+                    getattr(ds, k, None) == v for f in self.args.filter for k, v in [f.split('=', 1)]):
+                logger.info(f'Skipping {ds.SOPInstanceUID} due to no matching filter')
+                return
+            package_dicom(ds, anonymize=self.args.anonymize, compress=self.args.compress)
             dst = self.args.dst
             if dst:
                 src = self.args.src
@@ -351,7 +395,8 @@ def handle_store(event, handler):
 
 
 def extracted_list(path):
-    return {tuple(row) for row in csv.reader(open(path))}
+    if Path(path).is_file():
+        return {tuple(row) for row in csv.reader(open(path))}
 
 
 def get_ip():
@@ -367,10 +412,10 @@ def get_ip():
         s.close()
 
 
-def main():
+def read_args():
     parser = argparse.ArgumentParser('EchoLoader')
     parser.add_argument(
-        "--src", type=Path,
+        "--src",
         help="The folder to anonymize")
     parser.add_argument(
         "--dst",
@@ -387,6 +432,11 @@ def main():
     parser.add_argument(
         "--pacs-port", default=11112, type=int,
         help="PACS port, defaults to 11112")
+    parser.add_argument(
+        "--pacs-sops", type=str, nargs='+', default={
+            sop for sop in vars(sop_class).values()
+            if isinstance(sop, str) and sop.count('.') >= 5 and all(s.isdigit() for s in sop.split('.'))},
+        help="PACS accepted SOPs")
     parser.add_argument(
         "--overwrite", action="store_true",
         help="Overwrite files in the output folder")
@@ -397,10 +447,10 @@ def main():
         "--upload", action='store_true',
         help="Will upload all anonymized imaging to Us2.ai cloud")
     parser.add_argument(
-        "--env", type=ENV,
+        "--env",
         help="The Us2.ai environment to use")
     parser.add_argument(
-        "--extracted", type=extracted_list, default=set(),
+        "--extracted",
         help="File of cases to ignore - csv file with customer, trial, patientID, visit, filename")
     parser.add_argument(
         "--csv-out",
@@ -412,6 +462,10 @@ def main():
         '--no-anonymization', action='store_false', dest='anonymize',
         help="No anonymization of data prior to upload",
     )
+    parser.add_argument(
+        '--no-compression', action='store_false', dest='compress',
+        help="Don't compress videos into mp4 format",
+    )
     parser.add_argument(
         '--customer',
         help="The customer tag used for the upload (Admin only)",
@@ -420,7 +474,6 @@ def main():
         '--sync',
         nargs='+',
         default=[],
-        type=PacsConnection,
         help="Will try to sync structured report back to specified location. Format: IP:PORT:AE_TITLE",
     )
     parser.add_argument(
@@ -435,20 +488,85 @@ def main():
         action='store_true',
         help='Include main findings in SR export',
     )
+    parser.add_argument(
+        '--sync-designators',
+        nargs='+',
+        help='Designators used for SR export',
+    )
+    parser.add_argument(
+        '--sync-modalities',
+        default=['SR'],
+        nargs='+',
+        help='Modalities that should be synced to PACS, valid values are SR, PS, SC',
+    )
+    parser.add_argument(
+        '--sync-poll',
+        default=2,
+        type=float,
+        help='Delay between polls',
+    )
+    parser.add_argument(
+        '--username',
+        help='App username',
+    )
+    parser.add_argument(
+        '--password',
+        help='App password',
+    )
+    parser.add_argument(
+        '--config-file',
+        default=Path("config.json"),
+        type=Path,
+        help='Configuration file for the EchoLoader',
+    )
+    parser.add_argument(
+        '--trial',
+        help='Trial name of studies to upload',
+    )
+    parser.add_argument(
+        '--proxy',
+        help='Spin up a proxy against env on the given port',
+        type=int,
+    )
+    parser.add_argument(
+        '--log',
+        help='File for logs',
+    )
+    parser.add_argument(
+        '--filter',
+        help='Inclusive filter for DS files to process',
+        default=[],
+        nargs='+',
+    )
     args = parser.parse_args(sys.argv[1:])
-    args.env = args.env or ENV('production')
+    if args.config_file.is_file():
+        with args.config_file.open() as f:
+            for k, v in json.load(f).items():
+                setattr(args, k, v)
+    args.env = (args.env and ENV(args.env)) or ENV('production')
+    args.src = args.src and Path(args.src)
+    args.extracted = (args.extracted and extracted_list(args.extracted)) or set()
+    args.sync = [PacsConnection(c) for c in args.sync]
     if args.verbose:
         logging.basicConfig()
         logging.getLogger().setLevel(logging.DEBUG)
-    if args.upload or args.sync:
+    if args.log:
+        fh = logging.FileHandler(args.log)
+        logging.getLogger().addHandler(fh)
+    if args.upload or args.sync or args.proxy:
         args.auth = Auth(
             args,
-            os.environ.get(f"{OS_VAR_PREFIX}COGNITO_USERNAME") or input("username: "),
-            os.environ.get(f"{OS_VAR_PREFIX}COGNITO_PASSWORD") or getpass("password: "),
+            args.username or os.environ.get(f"{OS_VAR_PREFIX}COGNITO_USERNAME") or input("username: "),
+            args.password or os.environ.get(f"{OS_VAR_PREFIX}COGNITO_PASSWORD") or getpass("password: "),
         )
         args.customer = args.customer or args.auth.customer()
+    return args
+
+
+def main():
+    args = read_args()
     handler = Handler(args)
-    sync = None
+    sync = pacs = httpd = None
     try:
         if args.src:
             src = args.src
@@ -460,21 +578,25 @@ def main():
             if args.watch:
                 logger.warning(f"watching folder {os.path.abspath(src)}")
                 src.mkdir(exist_ok=True, parents=True)
-                observer = args.observer = Observer()
+                observer = args.observer = PollingObserver()
                 observer.schedule(handler, src, recursive=True)
                 observer.start()
         if args.sync:
-            sync = Sync(args)
+            sync = Sync(args, handler.pool)
             sync.start()
         if args.pacs:
             logger.warning(
                 f"Starting pacs server on {get_ip()}:{args.pacs_port} with AE title {args.pacs_ae_title}")
             handlers = [(evt.EVT_C_STORE, handle_store, [handler])]
-            ae = AE()
-            ae.add_supported_context(UltrasoundMultiframeImageStorage, ALL_TRANSFER_SYNTAXES)
-            ae.add_supported_context(UltrasoundImageStorage, ALL_TRANSFER_SYNTAXES)
-            ae.start_server(('0.0.0.0', args.pacs_port), block=True, evt_handlers=handlers, ae_title=args.pacs_ae_title)
-        while hasattr(args, 'observer') or handler.processing() or sync:
+            ae = AE(args.pacs_ae_title)
+            for sop in args.pacs_sops:
+                ae.add_supported_context(sop, ALL_TRANSFER_SYNTAXES)
+            pacs = ae.start_server(('0.0.0.0', args.pacs_port), block=False, evt_handlers=handlers)
+        if args.proxy:
+            httpd = ThreadingHTTPServer(server_address=('0.0.0.0', args.proxy), RequestHandlerClass=create_proxy(args))
+            thread = threading.Thread(target=httpd.serve_forever)
+            thread.start()
+        while hasattr(args, 'observer') or handler.processing() or sync or pacs or httpd:
             time.sleep(1)
     except KeyboardInterrupt:
         logger.warning("Interrupted, finishing up jobs")
@@ -483,5 +605,8 @@ def main():
         to_wait = [e for e in to_wait if e]
         for e in to_wait:
             e.stop()
+        for server in [pacs, httpd]:
+            if server:
+                server.shutdown()
         for e in to_wait:
             e.join()
